@@ -283,7 +283,7 @@ static int convert_hex(const char *pattern, size_t len, size_t& pos, convert_fla
     size_t n = pos + 3;
     if (c == 'u')
       n += 2;
-    while (k < n && k < len && std::isxdigit(c = pattern[k++]))
+    while (k < n && k < len && std::isxdigit(static_cast<unsigned char>(c = pattern[k++])))
       *s++ = c;
     *s = '\0';
     --k;
@@ -337,7 +337,7 @@ static const std::string& expand(const std::map<std::string,std::string> *macros
 {
   // lookup {name} and expand without converting
   size_t k = pos++;
-  while (pos < len && (std::isalnum(pattern[pos]) || pattern[pos] == '_' || (pattern[pos] & 0x80) == 0x80))
+  while (pos < len && (std::isalnum(static_cast<unsigned char>(pattern[pos])) || pattern[pos] == '_' || (pattern[pos] & 0x80) == 0x80))
     ++pos;
   if (pos >= len || (pattern[pos] == '\\' ? pattern[pos + 1] != '}' : pattern[pos] != '}'))
     throw regex_error(regex_error::undefined_name, pattern, pos);
@@ -415,6 +415,7 @@ static void insert_list(const char *pattern, size_t len, size_t& pos, convert_fl
 
 static void expand_list(const char *pattern, size_t len, size_t& loc, size_t& pos, convert_flag_type flags, const std::map<size_t,std::string>& mod, const char *signature, const char *par, const std::map<std::string,std::string> *macros, std::string& regex, bool& nl)
 {
+  size_t bls = pos - 1;
   bool invert = false;
   bool newline = false;
   if (pos + 1 < len && pattern[pos] == '^')
@@ -497,12 +498,37 @@ static void expand_list(const char *pattern, size_t len, size_t& loc, size_t& po
       pos += 3;
       expand_list(pattern, len, loc, pos, flags, mod, signature, par, macros, regex, nl1);
     }
+    else if ((c & 0xC0) == 0xC0 && is_modified(mod, 'u'))
+    {
+      // Unicode normalization may need to combine a previous ASCII character with a Unicode combining character
+      if (loc < pos && std::isalpha(static_cast<unsigned char>(pattern[pos - 1])))
+        --pos;
+      regex.append(&pattern[loc], pos - loc);
+      const char *s = &pattern[pos];
+      c = utf8(s, &s);
+      while (true)
+      {
+        // try combining one or more Unicode combining characters, if any
+        pos = s - pattern;
+        if (pos >= len || (*s & 0xC0) != 0xC0)
+          break;
+        int next = utf8(s, &s);
+        int combined = Unicode::compose(c, next);
+        if (combined < 0)
+          break;
+        c = combined;
+      }
+      char buf[8];
+      size_t size = utf8(c, buf);
+      regex.append(buf, size);
+      loc = pos--;
+    }
     ++pos;
     if (pos >= len || pattern[pos] == ']')
       break;
   }
   if (pos >= len || pattern[pos] != ']')
-    throw regex_error(regex_error::mismatched_brackets, pattern, loc);
+    throw regex_error(regex_error::mismatched_brackets, pattern, bls);
   if (invert && !newline && (flags & convert_flag::notnewline))
   {
     regex.append(&pattern[loc], pos - loc).append("\\n");
@@ -525,7 +551,7 @@ static void insert_escape_class(const char *pattern, size_t pos, convert_flag_ty
     wc = Posix::range(name);
   if (wc == NULL)
     throw regex_error(regex_error::invalid_class, pattern, pos);
-  if (std::islower(c))
+  if (std::islower(static_cast<unsigned char>(c)))
   {
     if (wc[0] <= '\n' && wc[1] >= '\n' && (flags & convert_flag::notnewline))
     {
@@ -719,7 +745,7 @@ static int insert_escape(const char *pattern, size_t len, size_t& pos, convert_f
     pos = k;
     return -1;
   }
-  else if (std::isalpha(c))
+  else if (std::isalpha(static_cast<unsigned char>(c)))
   {
     const char *s = std::strchr(regex_abtnvfr, c);
     if (s == NULL)
@@ -991,12 +1017,24 @@ static void insert_list(const char *pattern, size_t len, size_t& pos, convert_fl
     }
     else
     {
-      if ((c & 0xC0) == 0xC0 && is_modified(mod, 'u'))
+      if (is_modified(mod, 'u'))
       {
-        // unicode: UTF-8 sequence
-        const char *r;
-        c = utf8(&pattern[pos], &r);
-        pos += r - &pattern[pos] - 1;
+        // Unicode normalization
+        const char *s = &pattern[pos];
+        c = utf8(s, &s);
+        while (true)
+        {
+          // try combining one or more Unicode combining characters, if any
+          pos = s - pattern;
+          if (pos >= len || (*s & 0xC0) != 0xC0)
+            break;
+          int next = utf8(s, &s);
+          int combined = Unicode::compose(c, next);
+          if (combined < 0)
+            break;
+          c = combined;
+        }
+        --pos;
       }
       if (range)
       {
@@ -1023,12 +1061,12 @@ static void insert_list(const char *pattern, size_t len, size_t& pos, convert_fl
     }
   }
   if (pos >= len || pattern[pos] != ']')
-    throw regex_error(regex_error::mismatched_brackets, pattern, loc);
+    throw regex_error(regex_error::mismatched_brackets, pattern, loc - 1);
   if (negate)
     negate_list(flags, mod, ranges);
   extend_list(pattern, len, pos, flags, mod, ranges, macros);
   if (ranges.empty())
-    throw regex_error(regex_error::empty_class, pattern, loc);
+    throw regex_error(regex_error::empty_class, pattern, loc - 1);
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -1056,7 +1094,7 @@ static void convert_escape_char(const char *pattern, size_t len, size_t& loc, si
   else if (std::strchr(regex_meta, c) == NULL)
   {
     char buf[3] = { '^', static_cast<char>(lowercase(c)), '\0' };
-    bool invert = std::isupper(c) != 0;
+    bool invert = std::isupper(static_cast<unsigned char>(c)) != 0;
     if (c == 'n' || (invert && strchr("DHLUWX", c) != NULL))
       nl = true;
     const char *name = buf + !invert;
@@ -1248,7 +1286,7 @@ static void convert_escape(const char *pattern, size_t len, size_t& loc, size_t&
       throw regex_error(regex_error::invalid_escape, pattern, pos);
     if (wc == '\n')
       nl = true;
-    if (std::isalpha(wc) && is_modified(mod, 'i'))
+    if (std::isalpha(static_cast<unsigned char>(wc)) && is_modified(mod, 'i'))
     {
       // anycase: translate A to [Aa]
       regex.append(&pattern[loc], pos - loc - 1).push_back('[');
@@ -1297,7 +1335,7 @@ static void convert_escape(const char *pattern, size_t len, size_t& loc, size_t&
         if (wc <= 0xFF)
         {
           // translate \u{X}, \u00XX (convert_flag::u4) and \x{X} to \xXX
-          if (std::isalpha(wc) && is_modified(mod, 'i'))
+          if (std::isalpha(static_cast<unsigned char>(wc)) && is_modified(mod, 'i'))
           {
             // anycase: translate A to [Aa]
             regex.append(&pattern[loc], pos - loc - 1).push_back('[');
@@ -1414,7 +1452,7 @@ static void convert_escape(const char *pattern, size_t len, size_t& loc, size_t&
     }
     loc = pos + 1;
   }
-  else
+  else if (c != ' ' && c != '\t')
   {
     convert_escape_char(pattern, len, loc, pos, flags, signature, mod, par, regex, nl);
   }
@@ -1460,7 +1498,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
       mods.push_back('m');
     size_t k = 2;
     bool invert = false;
-    while (k < len && (pattern[k] == '-' || std::isalpha(pattern[k])))
+    while (k < len && (pattern[k] == '-' || std::isalpha(static_cast<unsigned char>(pattern[k]))))
     {
       if (pattern[k] == '-')
       {
@@ -1673,7 +1711,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
                 std::string mods, unmods;
                 size_t k = pos;
                 bool invert = false;
-                while (k < len && (pattern[k] == '-' || std::isalnum(pattern[k])))
+                while (k < len && (pattern[k] == '-' || std::isalnum(static_cast<unsigned char>(pattern[k]))))
                 {
                   if (pattern[k] == '-')
                   {
@@ -1968,7 +2006,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
         }
         else
         {
-          if (macros != NULL && pos + 1 < len && (std::isalpha(pattern[pos + 1]) || pattern[pos + 1] == '_' || pattern[pos + 1] == '$' || (pattern[pos + 1] & 0x80) == 0x80))
+          if (macros != NULL && pos + 1 < len && (std::isalpha(static_cast<unsigned char>(pattern[pos + 1])) || pattern[pos + 1] == '_' || pattern[pos + 1] == '$' || (pattern[pos + 1] & 0x80) == 0x80))
           {
             // if macros are provided: lookup {name} and expand without converting
             regex.append(&pattern[loc], pos - loc);
@@ -2001,10 +2039,11 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
             if (beg)
               throw regex_error(regex_error::empty_expression, pattern, pos);
             ++pos;
-            if (pos >= len || !std::isdigit(pattern[pos]))
+            if (pos >= len || !std::isdigit(static_cast<unsigned char>(pattern[pos])))
               throw regex_error(regex_error::invalid_repeat, pattern, pos);
             char *s;
             size_t n = static_cast<size_t>(std::strtoul(&pattern[pos], &s, 10));
+            size_t brc = pos - 1;
             pos = s - pattern;
             if (pos + 1 < len && pattern[pos] == ',')
             {
@@ -2031,7 +2070,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
               if (pos + 1 < len)
                 throw regex_error(regex_error::invalid_repeat, pattern, pos);
               else
-                throw regex_error(regex_error::mismatched_braces, pattern, pos);
+                throw regex_error(regex_error::mismatched_braces, pattern, brc);
             }
             if (pos + 1 < len && (pattern[pos + 1] == '?' || pattern[pos + 1] == '+') && !supports_escape(signature, pattern[pos + 1]))
               throw regex_error(regex_error::invalid_quantifier, pattern, pos + 1);
@@ -2165,7 +2204,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
         beg = false;
         break;
       default:
-        if (std::isalpha(pattern[pos]))
+        if (std::isalpha(static_cast<unsigned char>(pattern[pos])))
         {
           if (is_modified(mod, 'i'))
           {
@@ -2177,26 +2216,44 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
             loc = pos + 1;
           }
         }
-        else if ((c & 0xC0) == 0xC0 && is_modified(mod, 'u') && !supports_escape(signature, 'p'))
+        else if ((c & 0xC0) == 0xC0 && is_modified(mod, 'u'))
         {
-          // unicode: group UTF-8 sequence
+          // Unicode normalization may need to combine a previous ASCII character with a Unicode combining character
+          if (loc < pos && std::isalpha(static_cast<unsigned char>(pattern[pos - 1])))
+            --pos;
           regex.append(&pattern[loc], pos - loc);
-          loc = pos;
-          while (pos + 1 < len && ((c = pattern[++pos]) & 0xC0) == 0x80)
-            continue;
+          const char *s = &pattern[pos];
+          c = utf8(s, &s);
+          while (true)
+          {
+            // try combining one or more Unicode combining characters, if any
+            pos = s - pattern;
+            if (pos >= len || (*s & 0xC0) != 0xC0)
+              break;
+            int next = utf8(s, &s);
+            int combined = Unicode::compose(c, next);
+            if (combined < 0)
+              break;
+            c = combined;
+          }
+          char buf[8];
+          size_t size = utf8(c, buf);
           if (pos < len &&
+              !supports_escape(signature, 'p') &&
               (pattern[pos] == '*' ||
                ((flags & convert_flag::basic) && pos + 1 < len ?
                 (pattern[pos] == '\\' && (pattern[pos + 1] == '?' || pattern[pos + 1] == '+' || pattern[pos + 1] == '{')) :
                 (pattern[pos] == '?' || pattern[pos] == '+' || pattern[pos] == '{'))))
           {
-            regex.append(par).append(&pattern[loc], pos - loc).push_back(')');
-            loc = pos;
-            --pos;
+            // group normalized Unicode character as UTF-8
+            regex.append(par).append(buf, size).push_back(')');
+            loc = pos--;
           }
           else if (pos > loc)
           {
-            --pos;
+            // save normalized Unicode character as UTF-8
+            regex.append(buf, size);
+            loc = pos--;
           }
         }
         anc = false;
