@@ -59,6 +59,9 @@
 
 #ifdef OS_WIN
 # include <io.h>
+# ifdef __CYGWIN__
+#  include <unistd.h>
+# endif
 #else
 # include <unistd.h>
 #endif
@@ -72,7 +75,7 @@
 #endif
 
 // DO NOT ALTER THIS LINE: the makemake.sh script updates the version
-#define REFLEX_VERSION "3.5.1"
+#define REFLEX_VERSION "4.1.2"
 
 /// RE/flex scanner generator class, a variation of the classic "lex" tool to generate scanners.
 /**
